import datetime
import logging
import os
from collections import defaultdict
from typing import Dict, Optional, Collection, Iterable, List

from sqlalchemy import or_, update, case, null, func, and_
from sqlalchemy.dialects.postgresql import insert

from index_registry.mongodb.leases_db_interface import DbManager as LeasesDbManager
from index_registry.sqs.sqs_queues import get_consumer_name
from index_registry.utils.gcs import get_matching_gs_keys
from index_registry.utils.db import ReadWriteDbPool
from index_registry.sql.models import Index, Searcher
from pscr_common.utils import get_utcnow_with_tzinfo, datetime_from_ms_timestamp, EPOCH

from index_registry.engine.base import BaseSerializableClass
from index_registry.engine.org_sqs_engine import OrgSqsEngine

from index_registry.utils.index_field_helpers import properties_from_key, \
    get_size_from_props, get_type_from_props

_logger = logging.getLogger(__name__)

DISABLE_NON_VDR_INDEXES_ENV = "DISABLE_NON_VDR_INDEXES"


class IndexResponse(BaseSerializableClass):
    def __init__(self, index_path: str, org_id: str, prefix: str, partition: str, retention_class: str, size_MB: int,
                 timestamp: datetime.datetime, last_search_time: Optional[datetime.datetime], s3_event_sequencer: str,
                 priority: Optional[int] = None, date_removed: Optional[datetime.datetime] = None,
                 last_modified: Optional[datetime.datetime] = None, expiration_date: Optional[datetime.datetime] = None):
        BaseSerializableClass.__init__(self, index_path=index_path, org_id=org_id, prefix=prefix, partition=partition,
                                       retention_class=retention_class,size_MB=size_MB, timestamp=timestamp,
                                       last_search_time=last_search_time, s3_event_sequencer=s3_event_sequencer,
                                       priority=priority, date_removed=date_removed, last_modified=last_modified,
                                       expiration_date=expiration_date)


class IndexEngine(object):
    org_retention_classes = None
    retention_defaults = None

    def __init__(self, trigger_suffix, config, db_pool: ReadWriteDbPool):
        self.trigger_suffix = trigger_suffix
        self.lease_table = LeasesDbManager(config)
        self.index_bucket = config.get('lease', 'gcs_index_bucket')
        self.org_sqs_engine = OrgSqsEngine(config)
        self.db_pool = db_pool
        self.stage_2_placeholder_idle_expiration_days = config.getint("searcher",
                                                                      "stage_2_placeholder_idle_expiration_days",
                                                                      fallback=7)
        self.enable_org_sqs = config.getboolean("sqs",
                                                "enable_org_sqs",
                                                fallback=False)
        self.env = config.get('env', 'id')

    @staticmethod
    def get_tstamp_from_props(props):
        tstamp = props.get('tstamp', '0')
        return datetime_from_ms_timestamp(int(tstamp, 16))

    @staticmethod
    def get_expiration_date_from_props(props, expiration_time):
        expiration_date = props.get('expire')
        if not expiration_date:
            return None

        try:
            date = datetime.datetime.strptime(expiration_date, '%Y-%m-%d')
            return expiration_time.replace(year=date.year, month=date.month, day=date.day)
        except Exception as e:
            _logger.error("Could not parse Index expire field! props=%s expiration_time=%s error=%s", props, expiration_time, e)
            return None

    @staticmethod
    def get_gs_uri(bucket, key):
        return 'gs://%s/%s' % (bucket, key)

    def get_dir_from_trigger(self, key):
        return key.rpartition(self.trigger_suffix)[0]

    @classmethod
    def tuple_from_uri(cls, uri, last_search_time=None, gcs_event_sequencer='', date_removed=None, priority=None, last_modified=None):
        props = properties_from_key(uri)
        index_size = get_size_from_props(props)
        prefix = props.get('type')
        partition = props.get('partition')
        org_id = props.get('org')
        timestamp = cls.get_tstamp_from_props(props)
        retention_class = props.get('class')
        expiration_date = cls.get_expiration_date_from_props(props, last_modified) if last_modified else None

        return IndexResponse(index_path=uri, org_id=org_id, prefix=prefix, partition=partition,
                             retention_class=retention_class, size_MB=index_size,
                             timestamp=timestamp, last_search_time=last_search_time, s3_event_sequencer=gcs_event_sequencer,
                             priority=priority, date_removed=date_removed, last_modified=last_modified, expiration_date=expiration_date)

    @classmethod
    def check_expired_from_uri(cls, uri, now=None, last_modified=EPOCH, stage_2_placeholder_idle_expiration_days=7):
        now = now if now else get_utcnow_with_tzinfo()
        props = properties_from_key(uri)

        expiration_date = cls.get_expiration_date_from_props(props, last_modified)
        if expiration_date:
            if expiration_date <= now:
                #_logger.warning("Encountered an index that is already expired based on expiration date - %s", uri)
                return True
            else:
                return False

        timestamp = cls.get_tstamp_from_props(props)
        type = get_type_from_props(props)
        if timestamp == EPOCH and type == 'merging':
            timestamp = last_modified  # Use last modified time if timestamp is 0, since this is merge placeholder
            retention_threshold = now - datetime.timedelta(days=stage_2_placeholder_idle_expiration_days)
        else:
            if os.environ.get(DISABLE_NON_VDR_INDEXES_ENV, "false").lower() == "true":
                _logger.warning("Encountered a non-VDR index in a VDR-only environment. Treating the index as expired.")
                return True
            else:
                _logger.warning("Encountered a non-VDR index. Treating the index as not expired.")
                return False
        # TODO: Note that retention is always provided in lower case for the class, while S3 has it camel-case
        # This is something we should fix consistently across projects

        if timestamp != EPOCH and timestamp < retention_threshold:
            #_logger.warning("Encountered an index that is already expired based on timestamp - %s", uri)
            return True
        else:
            return False

    def add_from_s3_uri(self, uri, s3_event_sequencer, last_modified=None, now=None):
        now = now if now else get_utcnow_with_tzinfo()
        if last_modified is None:
            last_modified = now
        date_removed = now if self.check_expired_from_uri(uri, now, last_modified, self.stage_2_placeholder_idle_expiration_days) else None

        index = self.tuple_from_uri(uri, gcs_event_sequencer=s3_event_sequencer, date_removed=date_removed, last_modified=last_modified)
        _logger.info("Add index: %s", index)

        self.add_indexes([index], update_gcs_event_sequencer=True)

    def get_all_active_index_paths(self):
        with self.db_pool.scoped_session() as session:
            return [i.index_path for i in
                    session.query(Index.index_path)
                        .filter(Index.date_removed == null())
                        .filter(or_(Index.prefix == 'optimized', Index.prefix == 'merging')).all()]

    def add_indexes(self, indexes, update_gcs_event_sequencer=False, dry_run=False, only_add_missing=False,
                    skip_new_leases_table=False):
        self._upsert_indexes(indexes, update_gcs_event_sequencer, dry_run=dry_run, only_add_missing=only_add_missing)
        if not dry_run and not skip_new_leases_table:
            for index in indexes:
                # Avoid special indexer:// prefix since those are used only for registrations, not leases
                # Todo: we probably do not need this check anymore since skip_new_leases_table from registration should
                # prevent us from ever seeing any indexer:// type paths, but leaving it for now and logging if happens.
                if index.date_removed is None and not index.index_path.startswith("indexer://"):
                    if index.prefix == "merging":
                        self.lease_table.add_index(self._index_to_key(index.index_path), True)
                    else:
                        self.lease_table.add_index(self._index_to_key(index.index_path), False)
                else:
                    _logger.warning(f"Found an index path trying to add to leases that starts with "
                                    f"indexer://: {index.index_path}")

    def _index_to_key(self, url):
        if "://" in url:
            return url  # already a fully qualified path
        else:
            # TODO this is likely to not be gs:// for GCS so we will need to figure this out later:
            # Note gs:// appears all over the place here, but should only be used for re-sync from S3
            # The GCS listener should just provide a proper URI.
            return f"gs://{self.index_bucket}/{url.replace(self.trigger_suffix, '')}"

    def delete_index(self, uri, s3_event_sequencer='', now=None, last_modified=None):
        now = now or get_utcnow_with_tzinfo()

        index = self.tuple_from_uri(uri, gcs_event_sequencer=s3_event_sequencer, date_removed=now, last_modified=last_modified)
        _logger.debug("Delete index: %s",  index)

        self._upsert_indexes([index], update_s3_event_sequencer=True if s3_event_sequencer else False)
        # Delete index from mongodb
        if index.prefix == 'merging':
            self.lease_table.remove_index(uri, True)
        else:
            self.lease_table.remove_index(uri, False)

    def delete_org_indexes(self, org_id, now=None):
        _logger.debug(f"Delete Org Indexes called for {org_id}")
        now = now or get_utcnow_with_tzinfo()

        org_indexes = self._get_indexes_for_org(org_id)

        active_indexes = [index_path for index_path, date_removed in org_indexes.items() if date_removed is None]

        self._upsert_indexes([self.tuple_from_uri(index_path, date_removed=now) for index_path in active_indexes])

        # Delete all removed indexes from mongodb
        for index_path in active_indexes:
            props = self.tuple_from_uri(index_path)
            if props.prefix == 'merging':
                self.lease_table.remove_index(index_path, True)
            else:
                self.lease_table.remove_index(index_path, False)

        return len(active_indexes)

    def _upsert_indexes(self, indexes, update_s3_event_sequencer=False, dry_run=False, only_add_missing=False):
        if only_add_missing:
            indexes_comprehension = [i.index_path for i in indexes]
            with self.db_pool.scoped_session() as session:
                db_rows = session.query(Index.index_path).filter(Index.index_path.in_(indexes_comprehension)).all()
            result_comprehension = [i.index_path for i in db_rows]
            diff = list(set(indexes_comprehension) - set(result_comprehension))
            if len(diff) > 0:
                _logger.info(f"Will update: {diff}")
            indexes = [i for i in indexes if i.index_path in diff]

        indexes.sort(key=lambda i: i.index_path)
        data = []
        for i in indexes:
            d = {'index_path': i.index_path, 'org_id': i.org_id,
                 'prefix': i.prefix, 'size_MB': i.size_MB, 'timestamp': i.timestamp,
                 'last_search_time': i.last_search_time, 's3_event_sequencer': i.s3_event_sequencer,
                 'partition': i.partition, 'retention_class': i.retention_class, 'date_removed': i.date_removed,
                 'last_modified': i.last_modified, 'expiration_date': i.expiration_date}
            if i.priority is not None:
                d['priority'] = i.priority
            data.append(d)

        if not dry_run and len(data) > 0:
            with self.db_pool.scoped_session() as session:
                insert_stmt = insert(Index)

                update_fields = {'org_id': insert_stmt.excluded.org_id, 'prefix': insert_stmt.excluded.prefix,
                                 'size_MB': insert_stmt.excluded.size_MB, 'timestamp': insert_stmt.excluded.timestamp,
                                 # Only update last_search_time and date_modified if incoming value is not null
                                 'last_search_time': case([(insert_stmt.excluded.last_search_time == null(),
                                                            Index.last_search_time)],
                                                          else_=insert_stmt.excluded.last_search_time),
                                 'partition': insert_stmt.excluded.partition,
                                 'retention_class': insert_stmt.excluded.retention_class,
                                 'expiration_date': Index.expiration_date}

                if update_s3_event_sequencer:
                    # Update request originated from s3 trigger
                    update_fields['s3_event_sequencer'] = insert_stmt.excluded.s3_event_sequencer
                    update_fields['date_removed'] = insert_stmt.excluded.date_removed
                    update_fields['last_modified'] = insert_stmt.excluded.last_modified
                    update_fields['expiration_date'] = insert_stmt.excluded.expiration_date
                else:
                    # Not updating s3_event_sequencer so this is not coming from an s3 trigger.
                    # Explicitly disallow updates to change to a non-null value of date_removed
                    # and noll value of date_modified when not coming from s3 trigger.
                    # This is required to allow searcher registrations to upsert without removing
                    # previously 'expired' indexes or clearing last_modified date.
                    update_fields['date_removed'] = case([(Index.date_removed == null(), insert_stmt.excluded.date_removed)],
                                                         else_=Index.date_removed)
                    update_fields['last_modified'] = case([(insert_stmt.excluded.last_modified == null(),
                                                            Index.last_modified)],
                                                          else_=insert_stmt.excluded.last_modified)

                upsert_stmt = insert_stmt.on_conflict_do_update(
                    index_elements=[Index.index_path],
                    set_=update_fields,
                    where=or_(Index.last_modified == null(), insert_stmt.excluded.last_modified == null(),
                              Index.last_modified <= insert_stmt.excluded.last_modified))

                session.execute(upsert_stmt, data)

    def expire_stale_index_partition(self, index_path, now=None) -> int:
        now = now if now else get_utcnow_with_tzinfo()

        index = self.tuple_from_uri(index_path)

        with self.db_pool.scoped_session() as session:
            stmt = update(Index).values(date_removed=now) \
                .where(Index.org_id == index.org_id) \
                .where(Index.prefix == index.prefix) \
                .where(Index.partition == index.partition) \
                .where(Index.timestamp == index.timestamp) \
                .where(Index.size_MB != index.size_MB) \
                .where(Index.date_removed == null()) \
                .returning(Index.index_path)
            results = session.execute(stmt)
            indexes = self._get_indexes_from_results(results)
            _logger.debug("Calling cleanup from stale_index_partition")
            sqs_delete_list = self._cleanup_sqs_from_index_results(indexes=indexes)
            ret = self._cleanup_leases_mongo_table_from_index_results(indexes)
            # Re-call in case an add happened while we were deleting from index lease table.
            self._cleanup_sqs_from_index_results(delete_indexes_list=sqs_delete_list)
            return ret

    @staticmethod
    def _add_legacy_retention_days(retentions):
        # legacy classes (None) use the 'default' retention class
        if 'default' in retentions:
            retentions[None] = retentions['default']

    def expire_indexes_with_expiration_date(self, now) -> int:
        conditions = [Index.date_removed == null(), Index.expiration_date != null(), Index.expiration_date <= now]
        limit = os.environ.get("EXPIRE_VDR_INDICES_LIMIT", 250)
        num_expired_indices = 0

        while True:
            with self.db_pool.scoped_session() as session:
                index_paths = [i.index_path for i in session.query(Index.index_path).filter(and_(*conditions)).limit(limit).all()]
                if not index_paths:
                    return num_expired_indices

                update_stmt = update(Index).values(date_removed=now).where(Index.index_path.in_(index_paths))
                session.execute(update_stmt)
                _logger.debug("Calling cleanup from expire_indexes_with_expiration_date")
                sqs_delete_list = self._cleanup_sqs_from_index_results(indexes=index_paths)
                num_expired_indices += self._cleanup_leases_mongo_table_from_index_results(index_paths)
                # Re-call in case an add happened while we were deleting from index lease table.
                self._cleanup_sqs_from_index_results(delete_indexes_list=sqs_delete_list)

    def expire_indexes(self, retentions_by_org: Dict[str, Dict[str, int]], retention_defaults: Dict[str, int],
                       now=None) -> int:
        # gather all the orgs and retentions so that we can find all the orgs that aren't set to use the default
        #  values for each retention class
        # Expire all the non-defaulted retention orgs
        # Expire all the indexes by class where the org is not in the non-defaulted list

        expired_index_count = 0
        now = now if now else get_utcnow_with_tzinfo()

        # legacy classes (None) use the 'default' retention class
        self._add_legacy_retention_days(retention_defaults)
        for retentions in retentions_by_org.values():
            self._add_legacy_retention_days(retentions)

        retention_orgs_by_day_by_class = self._get_retention_orgs_by_day_by_class(retentions_by_org,
                                                                                  retention_defaults.keys())

        for retention_class, default_days in retention_defaults.items():
            orgs_not_using_default_days = []

            # expire by class for orgs not using default retentions
            for days, orgs in retention_orgs_by_day_by_class.get(retention_class, {}).items():
                if days != default_days:
                    expired_index_count += self._expire_org_indexes_by_class(
                        orgs=orgs,
                        not_orgs=None,
                        retention_class=retention_class,
                        retention_threshold=now-datetime.timedelta(days=days),
                        now=now)
                    orgs_not_using_default_days.extend(orgs)

            # expire by class using default retention and org not included above
            #  This will catch any orgs not explicitly configured in org_retention service
            expired_index_count += self._expire_org_indexes_by_class(
                orgs=None,
                not_orgs=orgs_not_using_default_days,
                retention_class=retention_class,
                retention_threshold=now-datetime.timedelta(days=default_days),
                now=now)

        return expired_index_count

    def _expire_org_indexes_by_class(self, orgs: Optional[Collection[str]],
                                           not_orgs: Optional[Collection[str]],
                                           retention_class: str,
                                           retention_threshold: datetime.datetime,
                                           now: datetime.datetime) -> int:
        expired_index_count = 0
        expired_last = 1
        while expired_last > 0:
            expired_last = self._expire_org_indexes_by_class_chunk(
                orgs=orgs,
                not_orgs=not_orgs,
                retention_class=retention_class,
                retention_threshold=retention_threshold,
                now=now,
                limit=1000)
            expired_index_count += expired_last
        return expired_index_count

    def _expire_org_indexes_by_class_chunk(self, orgs: Optional[Collection[str]],
                                     not_orgs: Optional[Collection[str]],
                                     retention_class: str,
                                     retention_threshold: datetime.datetime,
                                     now: datetime.datetime,
                                     limit: int) -> int:
        # Remove chunk of indexes from the database as well as from DynamoDb, and remove any existing SQS Queues
        with self.db_pool.scoped_session() as session:
            conditions = [Index.timestamp != EPOCH, Index.timestamp <= retention_threshold,
                          Index.date_removed == null(), Index.expiration_date == null()]

            if retention_class:
                conditions.append(func.lower(Index.retention_class) == retention_class.lower())
            else:
                conditions.append(Index.retention_class == null())

            if orgs:
                conditions.append(Index.org_id.in_(orgs))
            if not_orgs:
                conditions.append(~Index.org_id.in_(not_orgs))

            indexes = session.query(Index.index_path).filter(and_(*conditions)).limit(limit).all()
            index_paths = [i.index_path for i in indexes]
            if not index_paths:
                return 0

            update_stmt = update(Index).values(date_removed=now).where(Index.index_path.in_(index_paths))
            session.execute(update_stmt)
            _logger.debug("Calling cleanup from _expire_org_indexes_by_class_chunk")
            sqs_delete_list = self._cleanup_sqs_from_index_results(index_paths)
            ret = self._cleanup_leases_mongo_table_from_index_results(indexes=index_paths)
            # Re-call in case an add happened while we were deleting from index lease table.
            self._cleanup_sqs_from_index_results(delete_indexes_list=sqs_delete_list)
            _logger.info(f'Expired {ret} indexes for threshold={retention_threshold}, '
                         f'retention_class={retention_class}, orgs={orgs}, not_orgs={not_orgs}')
            return ret

    def expire_stage_2_placeholders(self, now: datetime.datetime) -> int:
        retention_threshold = now - datetime.timedelta(days=self.stage_2_placeholder_idle_expiration_days)
        expired_count = 0
        while True:
            with self.db_pool.scoped_session() as session:
                stmt = session.query(Index.index_path) \
                    .filter(Index.timestamp == EPOCH) \
                    .filter(Index.prefix == 'merging') \
                    .filter(Index.last_modified != null()) \
                    .filter(Index.last_modified <= retention_threshold) \
                    .filter(Index.date_removed == null()) \
                    .order_by(Index.index_path) \
                    .limit(250) \
                    .with_for_update()

                update_stmt = update(Index) \
                    .values(date_removed=now) \
                    .where(Index.index_path.in_(stmt.scalar_subquery())) \
                    .returning(Index.index_path).execution_options(synchronize_session=False)

                results = session.execute(update_stmt)
                indexes = sorted(self._get_indexes_from_results(results))
                if not indexes:
                    break

                sqs_delete_list = self._cleanup_sqs_from_index_results(indexes=indexes)
                expired_count += self._cleanup_leases_mongo_table_from_index_results(indexes)
                # Re-call in case an add happened while we were deleting from index lease table.
                self._cleanup_sqs_from_index_results(delete_indexes_list=sqs_delete_list)

        return expired_count

    def _get_active_orgs(self, now: Optional[datetime.datetime]=None):
        now = now if now else get_utcnow_with_tzinfo()
        expiration_threshold = now - datetime.timedelta(hours=int(os.environ.get("STALE_ORG_THRESHOLD_HOURS", 72)))

        #  Get orgs with unleased non-raw indexes - treat these as active since we don't know times
        with self.db_pool.scoped_session(read_only=True) as session:
            query = session.query(Index.org_id.distinct()) \
                           .outerjoin(Searcher, and_(Searcher.index_path == Index.index_path,
                                                     Searcher.expiration > now)) \
                           .filter(Searcher.index_path == null()) \
                           .filter(Index.date_removed == null()) \
                           .filter(Index.prefix != 'raw')

            unregistered_orgs = {org_id[0] for org_id in query.all()}

        #  Get orgs with data more recent than expiration_threshold
        # select distinct on (indexes.org_id) indexes.org_id, endtime from indexes
        #  left outer join searchers on (indexes.index_path = searchers.index_path)
        #  where indexes.date_removed is null and searchers.expiration > now() and (endtime > now() - interval '5 day')
        #  order by org_id, endtime desc nulls last

        with self.db_pool.scoped_session(read_only=True) as session:
            query = session.query(Index.org_id.distinct()) \
                .outerjoin(Searcher, Searcher.index_path == Index.index_path) \
                .filter(Searcher.expiration > now) \
                .filter(Index.date_removed == null()) \
                .filter(Searcher.endtime > expiration_threshold) \

            recent_orgs = {org_id[0] for org_id in query.all()}

        return unregistered_orgs | recent_orgs

    def expire_stale_orgs(self, now=None) -> int:
        # A "stale" org is defined as an org that does not have an active raw partition
        # where active means we have a current registered searcher

        now = now if now else get_utcnow_with_tzinfo()

        active_orgs = self._get_active_orgs(now)

        with self.db_pool.scoped_session() as session:
            stmt = update(Index).values(date_removed=now) \
                .where(~Index.org_id.in_(active_orgs)) \
                .where(Index.prefix == 'optimized') \
                .where(Index.date_removed == null()) \
                .returning(Index.index_path)
            results = session.execute(stmt)
            indexes = self._get_indexes_from_results(results)
            _logger.debug("Calling cleanup from expire_stale_orgs")
            sqs_delete_list = self._cleanup_sqs_from_index_results(indexes)
            ret = self._cleanup_leases_mongo_table_from_index_results(indexes=indexes)
            # Re-call in case an add happened while we were deleting from index lease table.
            self._cleanup_sqs_from_index_results(delete_indexes_list=sqs_delete_list)
            return ret

    @staticmethod
    def _get_indexes_from_results(results) -> list:
        return [res[0] for res in results.fetchall()]

    def _cleanup_leases_mongo_table_from_index_results(self, indexes) -> int:
        del_count = 0
        for index in indexes:
            index_type = get_type_from_props(properties_from_key(index))
            if index_type == 'merging':
                self.lease_table.remove_index(index, True)
            else:
                self.lease_table.remove_index(index, False)
            del_count += 1
        return del_count

    def _cleanup_sqs_from_index_results(self, indexes=None, delete_indexes_list=None) -> list:
        _logger.debug("Cleaning up org sqs for expired index")
        if self.enable_org_sqs:
            if delete_indexes_list:
                self.org_sqs_engine.org_sqs_delete_batch(delete_indexes_list)
                return delete_indexes_list
            elif indexes:
                delete_indexes = []
                for index in indexes:
                    try:
                        properties = properties_from_key(index)
                        queue_name = get_consumer_name(properties, self.env)
                        delete_indexes.append((properties['org'], queue_name))
                    except KeyError:
                        _logger.error(f"Found bad entry {index} ignoring")
                self.org_sqs_engine.org_sqs_delete_batch(delete_indexes)
                # Return the delete_indexes list so that we can be called with it again and not have to rebuild it.
                return delete_indexes
            else:
                return []
        else:
            _logger.info("Org SQS is disabled")
            return []

    def purge_indexes(self, date_removed: datetime.datetime):
        with self.db_pool.scoped_session() as session:
            count = session.query(Index).filter(Index.date_removed != null(), Index.date_removed <= date_removed)\
                                        .delete(synchronize_session=False)
            return count

    @staticmethod
    def _get_retention_orgs_by_day_by_class(retentions_by_org: Dict[str, Dict[str, int]],
                                            default_retention_classes: Iterable[str]) -> Dict[str, Dict[int, List[str]]]:
        retention_class_days_orgs = defaultdict(lambda: defaultdict(list))

        for default_retention_class in default_retention_classes:
            for org, retentions in retentions_by_org.items():
                day = retentions.get(default_retention_class)
                if day is not None:
                    retention_class_days_orgs[default_retention_class][day].append(org)

        return retention_class_days_orgs

    def _get_indexes_for_org(self, org_id: str, protocol: str=None) -> Dict[str, bool]:
        with self.db_pool.scoped_session(read_only=True) as session:
            query = session.query(Index.index_path, Index.date_removed)\
                .filter(Index.org_id == org_id)

            if protocol:
                query = query.filter(Index.index_path.startswith(protocol))

            results = query.all()

            return {result.index_path: result.date_removed for result in results}

    def sync_from_s3(self, org_id: str, bucket: str) -> Dict[str, int]:
        _logger.debug("Sync S3 bucket %s for org %s", bucket, org_id)

        now = get_utcnow_with_tzinfo()

        s3_indexes = {}
        prefix = f"org={org_id}"

        current_indexes = self._get_indexes_for_org(org_id, "gs://")

        for key, last_modified in get_matching_gs_keys(bucket, prefix=prefix, suffix=self.trigger_suffix):
            index_path = self.get_gs_uri(bucket, self.get_dir_from_trigger(key))
            date_removed = now if self.check_expired_from_uri(index_path, now, last_modified, self.stage_2_placeholder_idle_expiration_days) else None

            index_tuple = self.tuple_from_uri(index_path, date_removed=date_removed, last_modified=last_modified)
            _logger.debug("Found %s", index_tuple)

            s3_indexes[index_path] = index_tuple

        s3_index_paths = set(s3_indexes.keys())

        missing_index_paths = s3_index_paths - set(current_indexes.keys())
        if missing_index_paths:
            self.add_indexes([s3_indexes[index_path] for index_path in missing_index_paths])

        current_active_indexes = {index_path for index_path, date_removed in current_indexes.items()
                                  if date_removed is None}
        old_index_paths = current_active_indexes - s3_index_paths
        old_indexes = []
        if old_index_paths:
            for index in old_index_paths:
                index_tuple = self.tuple_from_uri(index, date_removed=now)
                old_indexes.append(index_tuple)
                if index_tuple.prefix == 'merging':
                    self.lease_table.remove_index(index, True)
                else:
                    self.lease_table.remove_index(index, False)
        if len(old_indexes) > 0:
            self._upsert_indexes(old_indexes, update_s3_event_sequencer=False)

        return {"added": len(missing_index_paths), "removed": len(old_index_paths)}

