import logging
from index_registry.mongodb.orgsqs_db_interface import DbManager as OrgsqsDbManager, ItemDoesNotExist, IndexTimestamps
from index_registry.mongodb.orgsqs_query_entry import QueryEntry, QueryEntries
from index_registry.mongodb.leases_db_interface import DbManager as LeasesDbManager
from index_registry.utils.chunker import chunker


_logger = logging.getLogger(__name__)


class IndexLeaseEntryDoesNotExist(Exception):
    pass


class OrgSqsEngine:
    def __init__(self, config):
        #self.sqs_manager = SQSManager(config)
        self.org_sqs_table = OrgsqsDbManager(config)
        self.lease_table = LeasesDbManager(config)
        self.config = config
        # Todo:  This should go away with the SQS to RabbitMQ ticket
        self.sqs_queue_protect_minutes = config.getint("sqs",
                                                       "queue_protect_minutes",
                                                       fallback=60)

    def org_sqs_delete(self, org, queue_name):
        _logger.debug(f"Deleting {queue_name} from org sqs dynamodb table")
        self.org_sqs_table.delete_entry(org, queue_name)

    def org_sqs_delete_batch(self, delete_list):
        for chunk in chunker(delete_list, 1000):
            self.org_sqs_table.delete_entry_batch(chunk)

    def org_sqs_add(self, query: QueryEntry):
        self.org_sqs_table.add_entry(query)

    def org_sqs_add_bulk(self, entries: QueryEntries):
        self.org_sqs_table.add_entry_batch(entries)

    def org_sqs_retrieve(self, org_key: str, index_timestamps: IndexTimestamps):
        return self.org_sqs_table.get_entries(org_key, index_timestamps)

    def org_sqs_retrieve_count(self, org_key: str, index_timestamps: IndexTimestamps):
        return self.org_sqs_table.get_entries_count(org_key, index_timestamps)

    def org_sqs_add_if_valid_lease(self, query: QueryEntry):
        #  We have a race condition where it is possible for searchers to insert invalid entries into org/sqs
        #  for recently expired leases, attempt to catch those situation and refuse to insert into org/sqs
        try:
            # if we already have an entry in org/sqs then we can overwrite it.
            self.org_sqs_table.add_entry(query, insert_only_if_exists=True)
        except ItemDoesNotExist:
            # check if there is a valid lease
            if self.lease_table.entry_exists(query.index_path):
                self.org_sqs_table.add_entry(query)
                # Do a re-check just to make sure we didn't get caught in a race where sql purge deleted between
                # the check and the add, if so delete the entry.
                if not self.lease_table.entry_exists(query.index_path):
                    _logger.warning(f"Deleting org_sqs entry {query.consumer_name} because lease entry went away")
                    self.org_sqs_table.delete_entry(query.org_key, query.consumer_name)
                    raise IndexLeaseEntryDoesNotExist(f"Tried to add org sqs entry for {query.consumer_name} where "
                                                      f"lease entry did not exist")
            else:
                _logger.warning(f"Refused to add org/sqs entry for queue {query.consumer_name} "
                                f"because lease did not exist")
                raise IndexLeaseEntryDoesNotExist(f"Tried to add org sqs entry for {query.consumer_name} where lease"
                                                  f"entry did not exist")

    def org_sqs_raw_index_purge(self):
        raw_index_timeout_minutes = self.config.getint("org_sqs",
                                                       "raw_index_timeout_minutes",
                                                       fallback=5)
        self.org_sqs_table.delete_expired_raw(raw_index_timeout_minutes)

    def sns_purge(self, dry_run=False, wait=60):
        pass
        # Todo: There is a later ticket to change this, we can probably change this from a cronjob to
        # Just delete the stream in Rabbit mq when we delete the last consumer on the queue.
        # _logger.debug(f"Getting topics")
        # topics = self.sqs_manager.get_topics()
        # delete_count = 0
        # delete_after_waiting = 0
        # no_deleted_queues = []
        # for topic in topics:
        #     topic_arn = topic['TopicArn']
        #     topic_attributes = self.sqs_manager.get_topic_attributes(topic_arn)
        #     if topic_attributes.subscriptions_created == 0 \
        #             and topic_attributes.subscriptions_pending == 0:
        #         # We have a race where when we first create a topic, it can take a few seconds to create the first queue
        #         # and AWS isn't super fast about incrementing the subscriptions pending
        #         # So we want to see a subscription_deleted to show that we can delete the topic.
        #         # Due to another bug keep a list of the topics not deleted(see below)
        #         if topic_attributes.subscriptions_deleted > 0:
        #             if dry_run:
        #                 _logger.info(f"Topic {topic_arn} has no subscriptions and should be deleted")
        #             else:
        #                 self.sqs_manager.delete_topic(topic_arn)
        #                 delete_count += 1
        #         else:
        #             no_deleted_queues.append(topic_arn)
        #             _logger.debug(f"Topic found that never had subscriptions {topic['TopicArn']}")
        # # We have a list of topics that had no confirmed or deleted subscriptions, this shouldn't happen except
        # # AWS has a bug where they do not increase the deleted subscriptions when they are supposed to
        # # AWS ticket 9820154061 in the "Titan" account.
        # # What we are going to do is loop back through these and check again and delete if subscription
        # # confirmed is still zero, the Race condition should clear in a few seconds, by creating a
        # # queue after topic creation.
        # time.sleep(wait)
        # for topic_arn in no_deleted_queues:
        #     topic_attributes = self.sqs_manager.get_topic_attributes(topic_arn)
        #     if topic_attributes.subscriptions_created == 0 \
        #             and topic_attributes.subscriptions_pending == 0:
        #         if dry_run:
        #             _logger.info(f"Topic {topic_arn} has no subscriptions and should be deleted")
        #         else:
        #             self.sqs_manager.delete_topic(topic_arn)
        #             delete_after_waiting += 1
        # _logger.info(f"Deleted: {delete_count} Delayed Delete: {delete_after_waiting} Total "
        #              f"{delete_count + delete_after_waiting} SNS topics with no subscriptions")

    # def _build_subscription_map(self):
    #     # Todo:  This should go away with the SQS to RabbitMQ ticket
    #     _logger.info("Building Subscriptions Map")
    #     subscriptions_map = dict()
    #     for subscription in self.sqs_manager.get_topic_subscriptions():
    #         subscriptions_map[subscription['Endpoint'].split(':')[-1]] = subscription['SubscriptionArn']
    #     return subscriptions_map

    def sqs_queue_purge(self, dry_run=False):
        pass
        # Todo:  This should go away with the SQS to RabbitMQ ticket
        # found = 0
        # missing = 0
        # recently_created = 0
        #
        # subscriptions_map = self._build_subscription_map()
        # for queue_url in self.sqs_manager.queues():
        #     queue = self.sqs_manager.get_queue_from_url(queue_url)
        #     sqs_entry = self.org_sqs_table.get_entry(get_org_from_queue_name(queue), queue)
        #     if not sqs_entry:
        #         now = int(time.time())
        #         # If a queue is recently created do not delete, we have a race where it can take time for entries for a
        #         # newly created queue to be made in the org/sqs table.
        #         if self.sqs_manager.get_queue_create_timestamp(queue_url) < now - (self.sqs_queue_protect_minutes * 60):
        #             missing += 1
        #             if dry_run:
        #                 _logger.info(f"SQS Queue {queue} is orphaned and should be deleted")
        #             else:
        #                 if queue in subscriptions_map:
        #                     _logger.debug(f"SQS/SNS Subscription {subscriptions_map[queue]} is being deleted")
        #                     self.sqs_manager.unsubscribe_queue(subscriptions_map[queue])
        #                 else:
        #                     _logger.warning(f"Queue {queue} wasn't subscribed when trying to delete queue")
        #                 _logger.debug(f"SQS Queue {queue} is being deleted")
        #                 self.sqs_manager.delete_queue(queue_url)
        #         else:
        #             recently_created += 1
        #     else:
        #         found += 1
        # _logger.info(f"SQS Queues: Found {found}, Orphaned {missing}, Recently Created {recently_created}")
        # if not dry_run:
        #     _logger.info(f"Deleted {missing} SQS queues without entries in org/sqs dynamodb table")
        # else:
        #     _logger.info(f"SQS Queues: Found {found}, Orphaned {missing}")