from collections import defaultdict
import datetime
from hashlib import md5
import logging
from typing import Collection, Optional

from sqlalchemy.dialects.postgresql import insert
from sqlalchemy import and_, or_, func, null
from sqlalchemy.sql.expression import label, desc

from index_registry.engine.index_engine import IndexEngine
from index_registry.utils.index_field_helpers import properties_from_key
from index_registry.utils.config import Config
from index_registry.utils.db import ReadWriteDbPool
from index_registry.sql.models import Searcher, Index
from index_registry.engine.base import BaseSerializableClass
from pscr_common.utils import get_utcnow_with_tzinfo, datetime_to_ms_timestamp

_logger = logging.getLogger(__name__)


class SearcherRequest(BaseSerializableClass):
    def __init__(self, searcher_id: str, org_id: str, container_version: str, endpoint: str, starttime: datetime.datetime,
                 endtime: datetime.datetime, index_path: str, last_search_time: datetime.datetime, index_partition_id: str,
                 index_stage: int):
        BaseSerializableClass.__init__(self, searcher_id=searcher_id, org_id=org_id, container_version=container_version, endpoint=endpoint,
                                       starttime=starttime, endtime=endtime, index_path=index_path, last_search_time=last_search_time,
                                       index_partition_id=index_partition_id, index_stage=index_stage)


class SearcherResponse(BaseSerializableClass):
    def __init__(self, searcher_id: str, org_id: str, container_version: str, endpoint: str, starttime: int, endtime: int,
                 index_path: str, index_partition_id: str, index_stage: int, expiration: int, retention_class: str):
        BaseSerializableClass.__init__(self, searcher_id=searcher_id, org_id=org_id, container_version=container_version, endpoint=endpoint,
                                       starttime=starttime, endtime=endtime, index_path=index_path, index_partition_id=index_partition_id,
                                       index_stage=index_stage, expiration=expiration, retention_class=retention_class)

    @staticmethod
    def from_request(request: SearcherRequest, expiration: datetime.datetime):
        return SearcherResponse(searcher_id=request.searcher_id, org_id=request.org_id,
                                container_version=request.container_version, endpoint=request.endpoint,
                                starttime=request.starttime and datetime_to_ms_timestamp(request.starttime),
                                endtime=request.endtime and datetime_to_ms_timestamp(request.endtime),
                                index_path=request.index_path,
                                index_partition_id=request.index_partition_id,
                                index_stage=request.index_stage,
                                expiration=expiration and datetime_to_ms_timestamp(expiration),
                                retention_class=IndexEngine.tuple_from_uri(request.index_path).retention_class)


class TimeBoundResponse(BaseSerializableClass):
    def __init__(self, lower: int, upper: int):
        BaseSerializableClass.__init__(self, lower=lower, upper=upper)


class SearcherEngine(object):
    DEFAULT_EXPIRATION_SECONDS = 60
    DEFAULT_STAGE_0_EXPIRATION_THRESHOLD_MINUTES = 10
    PLACEHOLDER_EXPIRATION_SECONDS = 60 * 30
    PLACEHOLDER_CONTAINER_VERSION = 'placeholder'
    PLACEHOLDER_ENDPOINT = 'placeholder'

    def __init__(self, config: Config, db_pool: ReadWriteDbPool):
        self.config = config
        self.db_pool = db_pool
        self.index_engine = IndexEngine(".trigger", config, db_pool=db_pool)

    def register_searcher(self, searcher_id, org_id, container_version, endpoint,
                          start_time, end_time, index_path, last_search_time, index_partition_id, index_stage,
                          now=None, expiration_seconds=None):
        return self.register_searchers([SearcherRequest(searcher_id, org_id, container_version, endpoint,
                                                        start_time, end_time, index_path, last_search_time,
                                                        index_partition_id, index_stage)], now, expiration_seconds)[0]

    def register_placeholder_searcher(self, index_path, now=None):
        props = properties_from_key(index_path)
        generated_searcher_id = f"placeholder_{md5(index_path.encode('utf-8'), usedforsecurity=False).hexdigest()}"

        # Don't need to register a placeholder if there is already a searcher registered for this index.
        if not self.get_index_searcher(index_path, now):
            if props['type'] == 'optimized':
                expiration_seconds = self.config.getint("searcher", "stage_2_placeholder_expiration_sec")
            else:
                expiration_seconds = self.config.getint("searcher", "stage_1_placeholder_expiration_sec")

            # Null start/end times mean this searcher will not be returned as available but will not result in an "offline" searcher
            self.register_searcher(generated_searcher_id, props["org"], self.PLACEHOLDER_CONTAINER_VERSION,
                                   self.PLACEHOLDER_ENDPOINT, None, None, index_path, None, props["partition"],
                                   None, now, expiration_seconds)

    def register_searchers(self, searchers: Collection[SearcherRequest], now=None, expiration_seconds=None):
        _logger.debug("register_searchers: %s", searchers)
        now = now or get_utcnow_with_tzinfo()
        if expiration_seconds:
            expiration = now + datetime.timedelta(seconds=expiration_seconds)
        else:
            expiration = now + datetime.timedelta(seconds=self.config.getint("searcher", "expiration_seconds"))
        raw_index_expiration = now + datetime.timedelta(seconds=self.config.getint("searcher",
                                                                                   "raw_expiration_seconds"))

        data = [{'searcher_id': searcher.searcher_id, 'org_id': searcher.org_id,
                 'container_version': searcher.container_version, 'endpoint': searcher.endpoint,
                 'starttime': searcher.starttime, 'endtime': searcher.endtime,
                 'index_path': searcher.index_path, 'index_partition_id': searcher.index_partition_id,
                 'index_stage': searcher.index_stage,
                 'expiration': expiration if searcher.index_stage != 0 else raw_index_expiration}
                for searcher in searchers]

        with self.db_pool.scoped_session() as session:
            # Make sure the index path exists in the index dir and last_search_time is up-to-date
            # Setting skip_new_leases_table True here to prevent this path from attempting to add the lease to the
            # mongodb Leasing Table
            self.index_engine.add_indexes([self.index_engine.tuple_from_uri(searcher.index_path,
                                                                            last_search_time=searcher.last_search_time)
                                           for searcher in searchers],
                                          skip_new_leases_table=True)

            insert_stmt = insert(Searcher)

            upsert_stmt = insert_stmt.on_conflict_do_update(
                index_elements=[Searcher.org_id, Searcher.searcher_id, Searcher.container_version],
                set_={'endpoint': insert_stmt.excluded.endpoint, 'starttime': insert_stmt.excluded.starttime,
                      'endtime': insert_stmt.excluded.endtime, 'index_path': insert_stmt.excluded.index_path,
                      'index_partition_id': insert_stmt.excluded.index_partition_id,
                      'index_stage': insert_stmt.excluded.index_stage, 'expiration': insert_stmt.excluded.expiration})

            session.execute(upsert_stmt, data)

        _logger.info("Registered %s searchers: %s",
                     len(searchers), ["%s: %s" % (s.searcher_id, s.index_path) for s in searchers])

        return [SearcherResponse.from_request(s, expiration if s.index_stage != 0 else raw_index_expiration) for s in searchers]

    def get_searchers(self, org_id, start_time, end_time, index_path_filters=None, now=None, index_partition_ids=None):
        _logger.debug("get_searchers: %s %s - %s", org_id, start_time, end_time)
        # return all searchers with matching org id and search window falling
        #  anywhere within the searchers index time window
        #    eg search request start time is between start and end times or
        #         request end time is between start and end times or
        #         request start time is before start and request end time is after end
        #  left join with Searchers so that we get an empty result for any index that doesn't have an active searcher
        now = now or get_utcnow_with_tzinfo()
        with self.db_pool.scoped_session(read_only=True) as session:
            query = session.query(Index.org_id, Index.index_path, Index.retention_class,
                                  Searcher.searcher_id, Searcher.container_version,
                                  Searcher.endpoint, Searcher.starttime, Searcher.endtime, Searcher.index_partition_id,
                                  Searcher.index_stage, Searcher.expiration) \
                .outerjoin(Searcher, and_(Searcher.index_path == Index.index_path, Searcher.expiration > now)) \
                .filter(Index.date_removed == null(), Index.org_id == org_id) \
                .filter(or_(Searcher.endpoint == null(),
                            and_(Searcher.starttime <= start_time, Searcher.endtime >= start_time).self_group(),
                            and_(Searcher.starttime <= end_time, Searcher.endtime >= end_time).self_group(),
                            and_(Searcher.starttime >= start_time, Searcher.endtime <= end_time).self_group()
                            )
                        ) \
                .filter(or_(Index.prefix != "raw", Searcher.endpoint != null())) \
                .order_by(Index.index_path, Searcher.expiration.desc().nullslast())\
                .distinct(Index.index_path)

            if index_path_filters:
                query = query.filter(Index.index_path.in_(index_path_filters))

            if index_partition_ids:
                query = query.filter(Index.partition.in_(index_partition_ids))

            searchers = [SearcherResponse(searcher_id=row.searcher_id, org_id=row.org_id,
                                          container_version=row.container_version, endpoint=row.endpoint,
                                          starttime=row.starttime and datetime_to_ms_timestamp(row.starttime),
                                          endtime=row.endtime and datetime_to_ms_timestamp(row.endtime),
                                          index_path=row.index_path,
                                          index_partition_id=row.index_partition_id,
                                          index_stage=row.index_stage,
                                          expiration=row.expiration and datetime_to_ms_timestamp(row.expiration),
                                          retention_class=row.retention_class)
                         for row in query.all()]

        _logger.debug("Return searchers: %s", searchers)
        return searchers

    def purge_searchers(self, expiration):
        with self.db_pool.scoped_session() as session:
            count = session.query(Searcher).filter(Searcher.expiration <= expiration).delete(synchronize_session=False)
            return count

    def get_searcher_time_bounds(self, org_id, now=None):
        now = now or get_utcnow_with_tzinfo()

        with self.db_pool.scoped_session(read_only=True) as session:
            data = session.query(Searcher.org_id,
                                 label('min_starttime', func.min(Searcher.starttime)),
                                 label('max_endtime', func.max(Searcher.endtime))) \
                .filter(and_(Searcher.org_id == org_id, Searcher.expiration > now)) \
                .group_by(Searcher.org_id).first()

        return TimeBoundResponse(lower=data.min_starttime and datetime_to_ms_timestamp(data.min_starttime),
                                 upper=data.max_endtime and datetime_to_ms_timestamp(data.max_endtime))

    def get_validation_searchers(self, org_id, start_time, end_time, now=None):
        now = now or get_utcnow_with_tzinfo()
        with self.db_pool.scoped_session(read_only=True) as session:
            data = session.query(Searcher) \
                .filter(and_(Searcher.org_id == org_id, Searcher.expiration > now)) \
                .filter(or_(Searcher.endpoint == null(),
                            and_(Searcher.starttime <= start_time, Searcher.endtime >= start_time).self_group(),
                            and_(Searcher.starttime <= end_time, Searcher.endtime >= end_time).self_group(),
                            and_(Searcher.starttime >= start_time, Searcher.endtime <= end_time).self_group()
                            )
                        ) \
                .order_by(Searcher.container_version, Searcher.index_stage, desc(Searcher.expiration)) \
                .distinct(Searcher.container_version, Searcher.index_stage) \
                .all()

            return [SearcherResponse(searcher_id=row.searcher_id, org_id=row.org_id,
                                     container_version=row.container_version, endpoint=row.endpoint,
                                     starttime=row.starttime and datetime_to_ms_timestamp(row.starttime),
                                     endtime=row.endtime and datetime_to_ms_timestamp(row.endtime),
                                     index_path=row.index_path,
                                     index_partition_id=row.index_partition_id,
                                     index_stage=row.index_stage,
                                     expiration=row.expiration and datetime_to_ms_timestamp(row.expiration),
                                     retention_class=self.index_engine.tuple_from_uri(row.index_path).retention_class)
                    for row in data]

    def get_validation_searchers_by_stage(self, stage: int, count: int):
        with self.db_pool.scoped_session(read_only=True) as session:
            data = session.query(Searcher) \
                .filter(Searcher.expiration > get_utcnow_with_tzinfo()) \
                .filter(Searcher.index_stage == stage) \
                .filter(Searcher.starttime != null()) \
                .filter(Searcher.endtime != null()) \
                .filter(Searcher.endpoint != null()) \
                .order_by(desc(Searcher.expiration)) \
                .limit(count)

            return [SearcherResponse(searcher_id=row.searcher_id, org_id=row.org_id,
                                     container_version=row.container_version, endpoint=row.endpoint,
                                     starttime=row.starttime and datetime_to_ms_timestamp(row.starttime),
                                     endtime=row.endtime and datetime_to_ms_timestamp(row.endtime),
                                     index_path=row.index_path,
                                     index_partition_id=row.index_partition_id,
                                     index_stage=row.index_stage,
                                     expiration=row.expiration and datetime_to_ms_timestamp(row.expiration),
                                     retention_class=self.index_engine.tuple_from_uri(row.index_path).retention_class)
                    for row in data]

    def find_index_gaps(self, start_time: datetime.datetime, end_time: datetime.datetime, min_gap: datetime.timedelta,
                        now: datetime.datetime=None):
        now = now or get_utcnow_with_tzinfo()
        min_gap = min_gap.total_seconds() * 1000

        searchers = defaultdict(list)
        with self.db_pool.scoped_session(read_only=True) as session:
            data = session.query(Searcher) \
                .filter(Searcher.expiration > now) \
                .filter(or_(and_(Searcher.starttime <= start_time, Searcher.endtime >= start_time).self_group(),
                            and_(Searcher.starttime <= end_time, Searcher.endtime >= end_time).self_group(),
                            and_(Searcher.starttime >= start_time, Searcher.endtime <= end_time).self_group()
                            )
                        ) \
                .all()

            for row in data:
                searchers[row.org_id].append(SearcherResponse(searcher_id=row.searcher_id, org_id=row.org_id,
                                                              container_version=row.container_version,
                                                              endpoint=row.endpoint,
                                                              starttime=row.starttime and datetime_to_ms_timestamp(row.starttime),
                                                              endtime=row.endtime and datetime_to_ms_timestamp(row.endtime),
                                                              index_path=row.index_path,
                                                              index_partition_id=row.index_partition_id,
                                                              index_stage=row.index_stage,
                                                              expiration=row.expiration and datetime_to_ms_timestamp(row.expiration),
                                                              retention_class=self.index_engine.tuple_from_uri(
                                                                  row.index_path).retention_class))

        gaps = []
        start_time = datetime_to_ms_timestamp(start_time)
        for org, searchers in searchers.items():
            # Setting this to gap check start time will check the first org index for a gap before it started
            last_start_end = (start_time, start_time)

            for searcher in sorted(searchers, key=lambda x: x.starttime):
                if searcher.endtime > last_start_end[1]:
                    # this start/end is not fully encapulated by last start/end
                    #  check gap between this start/end and last
                    if (searcher.starttime - last_start_end[1]) > min_gap:
                        gaps.append({"orgKey": org,
                                     "start": last_start_end[1],
                                     "end": searcher.starttime})

                last_start_end = (searcher.starttime, searcher.endtime)

        return gaps

    def get_index_searcher(self, index_path: str, now: datetime.datetime) -> Optional[SearcherResponse]:
        now = now or get_utcnow_with_tzinfo()
        with self.db_pool.scoped_session(read_only=True) as session:
            data = session.query(Searcher) \
                          .filter(Searcher.expiration > now) \
                          .filter(Searcher.index_path == index_path) \
                          .first()

            if data:
                return SearcherResponse(searcher_id=data.searcher_id, org_id=data.org_id,
                                        container_version=data.container_version, endpoint=data.endpoint,
                                        starttime=data.starttime and datetime_to_ms_timestamp(data.starttime),
                                        endtime=data.endtime and datetime_to_ms_timestamp(data.endtime),
                                        index_path=data.index_path,
                                        index_partition_id=data.index_partition_id,
                                        index_stage=data.index_stage,
                                        expiration=data.expiration and datetime_to_ms_timestamp(data.expiration),
                                        retention_class=self.index_engine.tuple_from_uri(
                                            data.index_path).retention_class)

        return None
