#!/usr/bin/env python3

import argparse
import datetime
from datetime import datetime as dt
import json
import logging
import os
import sys
import time
import pickle
import pytz
import boto3

from index_registry.engine.index_engine import IndexEngine
from index_registry.utils.index_field_helpers import properties_from_key
from index_registry.engine.searcher_engine import SearcherEngine
from index_registry.utils.config import Config
from index_registry.utils.db import ReadWriteDbPool
from index_registry.utils.gcs import get_matching_gs_keys, verify_expired_gcs_path
from index_registry.sql.migrate import db_upgrade, create_alembic_rev
from index_registry.sqs.sqs_queues import SQSManager, get_consumer_name, get_org_from_consumer_name
from index_registry.engine.org_sqs_engine import OrgSqsEngine
from index_registry.mongodb.orgsqs_db_interface import DbManager as OrgsqsDbmanager
from pscr_common.utils import chunker, get_utcnow_with_tzinfo, datetime_to_ms_timestamp
from index_registry.maintenance.restore_registry import restore

_logger = logging.getLogger(__name__)


class EnvDefault(argparse.Action):
    def __init__(self, envvar, required=True, default=None, **kwargs):
        if envvar and envvar in os.environ:
            default = os.environ[envvar]
        if required and default:
            required = False

        super(EnvDefault, self).__init__(default=default, required=required,
                                         **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        # If nargs is 0 then this is either store true/false or store const
        #  We want to store true or const if env var evaluates to true
        if self.nargs == 0:
            values = bool(values)
            if not isinstance(self.const, bool):
                values = self.const if values else values
        setattr(namespace, self.dest, values)


def initialize_db_schema(config):
    db_upgrade(config)


def initialize_registry_from_gcs(bucket, config, db_pool, now=None, dry_run=False, use_gcs_indexes_file=False,
                                 write_gcs_indexes_file=False, gcs_indexes_file_name="indexes.pickle",
                                 only_add_missing=False, only_remove_extraneous=False):
    if not use_gcs_indexes_file:
        engine = IndexEngine(".trigger", config, db_pool)
        now = now if now else get_utcnow_with_tzinfo()
        indexes = []
        for key, last_modified in get_matching_gs_keys(bucket, suffix=".trigger"):
            index_path = engine.get_gs_uri(bucket, engine.get_dir_from_trigger(key))
            date_removed = now if engine.check_expired_from_uri(index_path, now, last_modified=last_modified) else None
            if "org=_" not in index_path:
                indexes.append(engine.tuple_from_uri(index_path, gcs_event_sequencer='', date_removed=date_removed, last_modified=last_modified))
        if dry_run:
            _logger.info("Dry Run")
        if write_gcs_indexes_file:
            _logger.info("Writing Indexes Save File")
            with open(gcs_indexes_file_name, 'wb') as outfile:
                pickle.dump(indexes, outfile)
    else:
        _logger.info("Reading Indexes Save File")
        with open(gcs_indexes_file_name, 'rb') as infile:
            indexes = pickle.load(infile)

    add_indexes(indexes, config, db_pool, dry_run=dry_run, only_add_missing=only_add_missing, only_remove_extraneous=only_remove_extraneous)


def delete_sqs_entries_not_in_index_registry(config, dry_run):
    env_id = config.get('env', 'id')
    engine = IndexEngine(".trigger", config, None)
    needed_queues = set()
    org_sqs_table = OrgsqsDbmanager(config)
    org_sqs_engine = OrgSqsEngine(config)
    found = 0
    missing = 0
    lease_table = engine.lease_table

    needed_queues.update([get_consumer_name(properties_from_key(i), env_id)
                          for i in lease_table.iterate_all_index_paths()])

    for entry in org_sqs_table.iterate_all_entries():
        if entry.type == 'merging' or entry.type == 'optimized':
            queue = entry.consumer_name
            if queue not in needed_queues:
                missing += 1
                if dry_run:
                    print(f"SQS entry {queue} is orphaned and should be deleted")
                else:
                    print(f"SQS entry {queue} is being deleted")
                    org_sqs_engine.org_sqs_delete(get_org_from_consumer_name(queue), queue)
            else:
                found += 1
    print(f"SQS Entries:  Found {found}, Orphaned {missing}")


def delete_sqs_queues_not_in_sqs_table(config, dry_run):
    org_sqs_engine = OrgSqsEngine(config)
    print("Purging SQS Queues")
    org_sqs_engine.sqs_queue_purge(dry_run=dry_run)


def delete_topics_with_no_subscriptions(config, dry_run):
    org_sqs_engine = OrgSqsEngine(config)
    print("Purging SNS Topics")
    org_sqs_engine.sns_purge(dry_run=dry_run)


def delete_sqs_raw_index_entries(config):
    org_sqs_engine = OrgSqsEngine(config)
    print("Purging Raw Queues")
    org_sqs_engine.org_sqs_raw_index_purge()


def delete_sqs_subscriptions_without_queues(config, dry_run):
    sqs_manager = SQSManager(config)
    for subscription in sqs_manager.get_topic_subscriptions():
        queue = subscription['Endpoint'].split(':')[-1]
        if not sqs_manager.queue_exists(queue):
            if dry_run:
                print(f"Subscription {subscription['SubscriptionArn']} does not have matching queue {queue}")
            else:
                print(f"Subscription {subscription['SubscriptionArn']} does not have matching queue {queue}, deleting")
                sqs_manager.unsubscribe_queue(subscription['SubscriptionArn'])


def populate_mongodb_from_gcs(config, add_indexes, delete_indexes):
    engine = IndexEngine(".trigger", config, None)
    all_gcs_keys = set()
    added = 0
    expired_indexes = 0
    bucket_name = config.get('lease', 'gcs_index_bucket')
    #table = engine.lease_table.table
    for key, last_modified in get_matching_gs_keys(bucket_name, suffix=".trigger"):
        full_gs_path = engine.get_gs_uri(bucket_name, engine.get_dir_from_trigger(key))
        if IndexEngine.check_expired_from_uri(full_gs_path, last_modified=last_modified):
            expired_indexes += 1
            continue
        if not add_indexes:
            if not engine.lease_table.entry_exists(full_gs_path):
                props = properties_from_key(full_gs_path)
                age = IndexEngine.get_tstamp_from_props(props) - get_utcnow_with_tzinfo()
                if "org=_" in full_gs_path:
                    _logger.debug(f"Test Org Ignoring {full_gs_path}")
                else:
                    _logger.info(f"Missing index in MongoDB: {full_gs_path}, Age: {age}")
                    added += 1
        else:
            if "org=_" in full_gs_path:
                _logger.debug(f"Test Org Ignoring {full_gs_path}")
            else:
                _logger.info(f"Adding {full_gs_path} to dynamodb")
                props = engine.tuple_from_uri(full_gs_path)
                if props.prefix == 'merging':
                    added += engine.lease_table.add_index(full_gs_path, True)
                else:
                    added += engine.lease_table.add_index(full_gs_path, False)
        all_gcs_keys.add(full_gs_path)

    total_items = 0
    items_to_delete = []
    for index_path in engine.lease_table.iterate_all_index_paths():
        if index_path not in all_gcs_keys:
            items_to_delete.append(index_path)
        total_items += 1

    for i in items_to_delete:
        props = properties_from_key(i)
        age = IndexEngine.get_tstamp_from_props(props) - get_utcnow_with_tzinfo()
        _logger.info(f"Extraneous index in MongoDb: {i}, age: {age}")

    if delete_indexes:
        for index in items_to_delete:
            props = engine.tuple_from_uri(index)
            if props.prefix == 'merging':
                engine.lease_table.remove_index(index, True)
            else:
                engine.lease_table.remove_index(index, False)
    _logger.warning(f"{'Added' if add_indexes else 'Detected'} {added} missing items")
    _logger.warning(f"{'Cleaned up' if delete_indexes else 'Detected'} {len(items_to_delete)} "
                    f"extraneous items out of {total_items} total. Expired indexes found: {expired_indexes}")


def add_index_from_path(index_path, config, db_pool):
    index = IndexEngine.tuple_from_uri(index_path, gcs_event_sequencer='')
    add_indexes([index], config, db_pool=db_pool)


def default_indexes_from_config(config):
    indexes = []

    now = get_utcnow_with_tzinfo(truncate_to_ms=True)
    now_ms = datetime_to_ms_timestamp(now)

    for org_id, index_count in config.items():
        for x in range(int(index_count)):
            index_path = f'default://mock-bucket/class=default/org={org_id}/partition={x}/type=optimized/tstamp={now_ms:x}/size=51200/index'
            indexes.append(IndexEngine.tuple_from_uri(index_path, gcs_event_sequencer='', priority=-1))

    return indexes


def add_indexes(indexes, config, db_pool, dry_run=False, only_add_missing=False, only_remove_extraneous=False):
    engine = IndexEngine(".trigger", config, db_pool=db_pool)
    if not only_add_missing:
        # Get all index paths from the DB so that we can figure out which ones should be deleted
        db_indexes = engine.get_all_active_index_paths()
        diff = set(db_indexes) - set([i.index_path for i in indexes])
        if only_remove_extraneous:
            indexes = []
        now = get_utcnow_with_tzinfo()
        if len(diff):
            # Double check that index is really gone
            diff = [index_path for index_path in diff if verify_expired_gcs_path(index_path)]
            _logger.info(f"Found: {len(diff)} extraneous indexes in DB out of {len(db_indexes)}")
            for index_path in diff:
                indexes.append(engine.tuple_from_uri(index_path, gcs_event_sequencer='', date_removed=now))
    _logger.info(f"Will update {len(indexes)} indexes")
    for indexes_chunk in chunker(100, indexes):
        for index in indexes_chunk:
            _logger.info("Adding index: %s to RDS", index["index_path"])
        engine.add_indexes(indexes_chunk, update_gcs_event_sequencer=True, dry_run=dry_run,
                           only_add_missing=only_add_missing)


def find_gaps(config, db_pool, starttime, endtime, min_gap):
    engine = SearcherEngine(config, db_pool=db_pool)

    return engine.find_index_gaps(starttime, endtime, datetime.timedelta(minutes=min_gap))


def _get_config(args):
    config = Config.create()
    
    # Introducing a 5 sec sleep before the actual script
    # to avoid errors like 'Unable to find credentials' by
    # kube2iam
    time.sleep(5)

    config.set('db', 'host', args.db_host)
    config.set('db', 'port', str(args.db_port))
    config.set('db', 'name', args.db_name)

    config.set('db', 'user', args.db_user)
    config.set('db', 'password', args.db_password)
    config.set('db', 'password_secrets_manager_key', args.db_password_secrets_manager_key)

    config.set('mongo', 'host', args.mongo_host)
    config.set('mongo', 'port', str(args.mongo_port))
    config.set('mongo', 'db_name', args.mongo_db_name)
    config.set('mongo', 'direct_connect', str(args.mongo_direct_connect))

    config.set('mongo', 'user', args.mongo_user)
    config.set('mongo', 'password', args.mongo_password)

    config.set('db.readonly', 'enabled', 'False')

    config.set('lease', 'gcs_index_bucket', args.index_bucket_name)

    config.set('env', 'id', args.env_id)
    config.set('sqs', 'queue_empty_timeout', str(args.delete_sqs_queue_timeout))

    return config


def create_snapshot(rds_client, db_cluster_id):
    # trunk-ignore(ruff/D401)
    """
    Function generates a timestamp and attempts to create a snapshot
    of the Aurora cluster.

    :param rds_client:
    :param db_cluster_id:
    """

    now = dt.now()
    TIMESTAMP_FORMAT = '%Y-%m-%d-%H-%M'
    timestamp_format = now.strftime(TIMESTAMP_FORMAT)

    # Create the snapshot name
    snapshot_identifier = '{}-{}'.format(db_cluster_id, timestamp_format)

    # Attempt to snapshot Aurora cluster
    try:
        response = rds_client.create_db_cluster_snapshot(
            DBClusterSnapshotIdentifier=snapshot_identifier,
            DBClusterIdentifier=db_cluster_id,
            Tags=[{'Key': 'CreatedBy', 'Value': 'Snapshot Tool for Aurora'}, {
                'Key': 'CreatedOn', 'Value': timestamp_format}]
        )
        return response
    except Exception as e:
        _logger.error(f"Failed to create snapshot {snapshot_identifier} ({e})")


def purge_snapshots(rds_client, db_cluster_id):
    DELETE_THRESHOLD_IN_DAYS = 7
    MIN_TO_KEEP = 1
    currentDate = dt.now().date()
    # Assume we will have more pages
    markerExists = True
    # Initialize marker i.e. pagination token
    marker = ""

    print(f"*** Delete threshold is {DELETE_THRESHOLD_IN_DAYS} days ***")
    print(f"*** Minimum Snapshots to be kept is {MIN_TO_KEEP} ***")

    toPurge = []
    totalSnapshots = 0

    # Emulate do while loop
    while markerExists:
        snapshots = rds_client.describe_db_cluster_snapshots(
            SnapshotType="manual", MaxRecords=100, Marker=marker, DBClusterIdentifier=db_cluster_id
        )

        if "Marker" in snapshots:
            marker = snapshots["Marker"]
            print("There are more pages...")
        else:
            markerExists = False
            print("No more pages left...")

        for snapshot in snapshots["DBClusterSnapshots"]:
            totalSnapshots += 1
            _snapCreationDate = snapshot["SnapshotCreateTime"]
            snapCreationDate = _snapCreationDate.date()
            diff = currentDate - snapCreationDate
            if diff.days > DELETE_THRESHOLD_IN_DAYS:
                toPurge.append(snapshot["DBClusterSnapshotIdentifier"])
        # End For Loop
    # End While Loop

    totalEligibleForDeletion = totalSnapshots - MIN_TO_KEEP

    print(f"Total snapshots: {totalSnapshots}" )
    print(f"Elegible for deletion: {totalEligibleForDeletion}" )

    if totalEligibleForDeletion > 0:
        for snapshotId in toPurge:
            try:
                rds_client.delete_db_cluster_snapshot(
                        DBClusterSnapshotIdentifier=snapshotId
                    )
                print(
                        f"*** {snapshotId} deleted. Creation date older than {DELETE_THRESHOLD_IN_DAYS} days"
                    )
            except Exception as e:
                if "InvalidDBClusterSnapshotStateFault" in e.message:
                    print(f"Skipping {snapshotId} due to {e}")
                    continue
            totalEligibleForDeletion -=1
            if totalEligibleForDeletion <= 0:
                break
        # End For Loop
    else:
        print(f"Skipped purging as the number of snapshots is less than {MIN_TO_KEEP}")


def create_purge_snapshot(db_cluster_id, snapshot_db):
    rds_client = boto3.client('rds')

    if not snapshot_db:
        _logger.info(f"Will not create DB snapshot for {db_cluster_id}.")
    else:
        purge_snapshots(rds_client, db_cluster_id)
        snapshot = create_snapshot(rds_client, db_cluster_id)
        _logger.info(f"Snapshot Details: {snapshot}")


def build_parser():
    parser = argparse.ArgumentParser()

    parser.add_argument('-v', '--verbose', action="store_true", dest='verbose',
                        required=False, default=False,
                        help='Verbose logging')
    parser.add_argument('--populate-rds-from-gcs-bucket', action=EnvDefault, envvar='LEASE_POPULATE_RDS', dest='lease_populate_rds', required=False, default=False,
                        help='Read indexes from gcs bucket and populate rds db, cannot be run with mongo populate')
    parser.add_argument('--populate-lease-mongo-from-gcs-bucket', action=EnvDefault, envvar='LEASE_POPULATE_MONGO', dest='lease_populate_mongo', required=False, default=False,
                        help='Read indexes from gcs bucket and populate mongo db, cannot be run with rds populate')
    parser.add_argument('--index-gcs-bucket', action=EnvDefault, envvar='LEASE_GCS_INDEX_BUCKET',
                        dest='index_bucket_name', type=str, required=False,
                        help='Index GCS bucket name')
    parser.add_argument('--env-id', action=EnvDefault,
                        envvar='ENV_ID', dest='env_id',
                        type=str, required=False, default=None,
                        help='Specify Environment Name')
    parser.add_argument('--db-host', action=EnvDefault, envvar='DB_HOST', dest='db_host',
                        type=str, required=False,
                        help='PostgreSQL host')
    parser.add_argument('--db-port', action=EnvDefault, envvar='DB_PORT', dest='db_port',
                        type=int, required=False, default=5432,
                        help='PostgreSQL port')
    parser.add_argument('--db-name', action=EnvDefault, envvar='DB_NAME', dest='db_name',
                        type=str, required=False,
                        help='PostgreSQL database name')
    parser.add_argument('--db-cluster-id', action=EnvDefault, envvar='RDS_HOSTNAME', dest='db_cluster_id',
                        type=str, required=False,
                        help='PostgreSQL database cluster ID')
    parser.add_argument('--db-user', action=EnvDefault, envvar='DB_USER', dest='db_user',
                        type=str, required=False,
                        help='PostgreSQL username')
    parser.add_argument('--db-password', action=EnvDefault, envvar='DB_PASSWORD', dest='db_password',
                        type=str, required=False, default='',
                        help='PostgreSQL password')
    parser.add_argument('--db-snapshot', action="store_true", dest='snapshot_db',
                        required=False, default=False,
                        help='Enable db snapshot')
    parser.add_argument('--db-password-secrets-manager-key', action=EnvDefault, envvar='DB_PASSWORD_SECRETS_MANAGER_KEY',
                        dest='db_password_secrets_manager_key', type=str, required=False, default='',
                        help='PostgreSQL password from secrets manager')
    parser.add_argument('--mongo-host', action=EnvDefault, envvar='MONGO_HOST', dest='mongo_host',
                        type=str, required=False,
                        help='MongoDB host')
    parser.add_argument('--mongo-port', action=EnvDefault, envvar='MONGO_PORT', dest='mongo_port',
                        type=str, required=False, default=27017,
                        help='MongoDB port')
    parser.add_argument('--mongo-db-name', action=EnvDefault, envvar='MONGO_DB_NAME', dest='mongo_db_name',
                        type=str, required=False, default="index_registry",
                        help='MongoDB database name')
    parser.add_argument('--mongo-user', action=EnvDefault, envvar='MONGO_USER', dest='mongo_user',
                        type=str, required=False,
                        help='MongoDB username')
    parser.add_argument('--mongo-password', action=EnvDefault, envvar='MONGO_PASSWORD', dest='mongo_password',
                        type=str, required=False, default='',
                        help='MongoDB password')
    parser.add_argument('--mongo-direct-connect', action=EnvDefault, envvar='MONGO_DIRECT_CONNECT', dest='mongo_direct_connect', required=False, default=False,
                        help='Add Direct connect to mongo connection string, good for port forwards')
    parser.add_argument('--init-db', action=EnvDefault, envvar='INIT_DB', nargs=0, const=True, dest='init_db',
                        required=False, default=False,
                        help='Initialize db')
    parser.add_argument('--add-index', action=EnvDefault, envvar='ADD_INDEX', dest='add_index',
                        type=str, required=False, default=None,
                        help='Manually add an index to the registry.')
    parser.add_argument('--defaults-from-config', action=EnvDefault, envvar='DEFAULTS_FROM_CONFIG',
                        dest='defaults_from_config',
                        type=str, required=False, default=None,
                        help='Json default index configuration.  Map of org_ids to default index count.')
    parser.add_argument('-r', '--revision', action="store_true", dest='revision', required=False, default=False,
                        help='Create alembic revision from models')
    parser.add_argument('-m', '--message', type=str, dest='message', required=False,
                        help='Revision message')
    parser.add_argument('-g', '--gaps', action="store_true", dest='gaps', required=False, default=False,
                        help='Find per org index gaps')
    parser.add_argument('-s', '--start', type=str, dest='start', required=False,
                        help='Gap check UTC start time in ISO format. eg 2019-03-04T14:12:32Z')
    parser.add_argument('-e', '--end', type=str, dest='end', required=False,
                        help='Gap check UTC end time in ISO format. eg 2019-03-04T14:12:32Z')
    parser.add_argument('--min-gap', type=int, dest='min_gap', required=False, default=60,
                        help='Minimum gap in minutes to report during gap check')
    parser.add_argument('--gcs-populate-dry-run', action="store_true", dest='gcs_populate_dry_run', required=False, default=False,
                        help='Dry run for populate from gcs')
    parser.add_argument('--gcs-use-indexes-save-file', action="store_true", dest='use_gcs_indexes_file', required=False, default=False,
                        help='Use a previously saved indexes file')
    parser.add_argument('--gcs-write-indexes-save-file', action="store_true", dest='write_gcs_indexes_file', required=False, default=False,
                        help='Save an indexes file')
    parser.add_argument('--gcs-indexes-save-file-name', type=str, dest='gcs_indexes_file_name', required=False, default='indexes.pickle',
                        help='Indexes File Name')
    parser.add_argument('--gcs-only-add-missing', action="store_true", dest='only_add_missing', required=False, default=False,
                        help='Only upsert missing indexes')
    parser.add_argument('--gcs-only-remove-extraneous', action="store_true", dest='only_remove_extraneous', required=False, default=False,
                        help='Only remove indexes that no longer exist on gcs')
    parser.add_argument('--add-lease-mongo', action="store_true", dest='add_leases_mongo', required=False, default=False,
                        help='Add missing indexes to Leases MongoDB (when populate_mongo_from_gcs is set)')
    parser.add_argument('--delete-lease-mongo', action="store_true", dest='delete_leases_mongo', required=False, default=False,
                        help='Delete extraneous indexes to Leases MongoDB (when populate_mongo_from_gcs is set)')
    parser.add_argument('--delete-sqs', action="store_true", dest='delete_sqs', required=False, default=False,
                        help='Delete extraneous sqs db entries')
    parser.add_argument('--delete-sqs-queue-timeout', type=int, dest='delete_sqs_queue_timeout', required=False, default=30,
                        help='Timeout to wait for queue to empty')
    parser.add_argument('--delete-sqs-dry-run', action="store_true", dest='delete_sqs_dry_run', required=False, default=False,
                        help='Delete extraneous sqs db entries')
    parser.add_argument('--delete-sqs-queues', action=EnvDefault, envvar='SQS_QUEUE_PURGE', dest='delete_sqs_queues', required=False, default=False,
                        help='Delete extraneous sqs queues')
    parser.add_argument('--delete-sqs-queues-dry-run', action="store_true", dest='delete_sqs_queues_dry_run', required=False, default=False,
                        help='Delete extraneous sqs queues')
    parser.add_argument('--delete-sqs-queue-subscriptions', action="store_true", dest='delete_sqs_queue_subscriptions', required=False, default=False,
                        help='Delete extraneous sqs subscriptions')
    parser.add_argument('--delete-sqs-queues-subscriptions-dry-run', action="store_true", dest='delete_sqs_queue_subscriptions_dry_run', required=False, default=False,
                        help='Delete extraneous sqs subscriptions')
    parser.add_argument('--delete-sns-topics', action=EnvDefault, envvar='SNS_TOPIC_PURGE', dest='delete_sns_topics', required=False, default=False,
                        help='Delete extraneous sns topics')
    parser.add_argument('--delete-sns-topics-dry-run', action="store_true", dest='delete_sns_topics_dry_run', required=False, default=False,
                        help='Delete extraneous sns topics')
    parser.add_argument('--delete-sqs-raw-purge', action=EnvDefault, envvar='SQS_RAW_PURGE', dest='delete_sqs_raw_purge', required=False, default=False,
                        help='Delete expired sqs raw entries')
    parser.add_argument('--registry-restore-source-leases-table', help='source table for leases', type=str, required=False, default=False)
    parser.add_argument('--registry-restore-target-leases-table', help='target table for leases', type=str, required=False, default=False)
    parser.add_argument('--registry-restore-org-key', help='use valid org key', type=str, required=False, default=False)
    return parser


def main():
    parser = build_parser()
    args = parser.parse_args()

    if args.verbose:
        log_level = logging.DEBUG
    elif "LOG_LEVEL" in os.environ:
        log_level = os.getenv("LOG_LEVEL")
    else:
        log_level = logging.INFO

    logging.basicConfig(stream=sys.stdout, level=log_level,
                        format='%(asctime)s.%(msecs)03d (%(levelname)s) - %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S')

    logging.getLogger('botocore').setLevel(logging.WARN)
    logging.getLogger('boto3').setLevel(logging.WARN)
    logging.getLogger('urllib3').setLevel(logging.WARN)

    config = _get_config(args)

    if args.db_cluster_id:
        # Read cluster identifier from RDS_HOSTNAME
        db_cluster_id = args.db_cluster_id.split(".")[0]
        if args.snapshot_db:
            print("Will do snapshot. {}".format(args.snapshot_db))
            create_purge_snapshot(db_cluster_id=db_cluster_id, snapshot_db=args.snapshot_db)

    if args.revision:  # pragma: no cover
        initialize_db_schema(config)
        create_alembic_rev(config, args.message)

    if args.init_db:
        initialize_db_schema(config)

    if args.delete_sqs:
        if not args.env_id:
            print("Environment id must be specified with --env-id")
            exit(1)
        delete_sqs_entries_not_in_index_registry(config, dry_run=args.delete_sqs_dry_run)
        print("If SQS delete specified don't do anything else")
        exit()

    if args.delete_sqs_queues:
        if not args.env_id:
            print("Environment id must be specified with --env-id")
            exit(1)
        delete_sqs_queues_not_in_sqs_table(config, dry_run=args.delete_sqs_queues_dry_run)
        print("If SQS delete queues specified don't do anything else")
        exit()

    if args.delete_sqs_queue_subscriptions:
        if not args.env_id:
            print("Environment id must be specified with --env-id")
            exit(1)
        delete_sqs_subscriptions_without_queues(config, dry_run=args.delete_sqs_queue_subscriptions_dry_run)
        print("If SQS delete queue subscription specified don't do anything else")
        exit()

    if args.delete_sns_topics:
        if not args.env_id:
            print("Environment id must be specified with --env-id")
            exit(1)
        delete_topics_with_no_subscriptions(config, dry_run=args.delete_sns_topics_dry_run)
        print("If SQS delete topics specified don't do anything else")
        exit()

    if args.delete_sqs_raw_purge:
        if not args.env_id:
            print("Environment id must be specified with --env-id")
            exit(1)
        delete_sqs_raw_index_entries(config)
        print("If SQS raw purge specified don't do anything else")
        exit()

    if args.registry_restore_org_key:
        if not args.registry_restore_source_leases_table:
            print("source leases table must be specified with "
                  "--registry-restore-source-leases-table")
            exit(1)
        if not args.registry_restore_target_leases_table:
            print("target leases table must be specified with "
                  "--registry-restore-target-leases-table")
            exit(1)
        restore(args, config)
        exit()

    if args.lease_populate_mongo:
        populate_mongodb_from_gcs(config, add_indexes=args.add_leases_mongo, delete_indexes=args.delete_leases_mongo)
    elif args.lease_populate_rds:
        with ReadWriteDbPool(config) as db_pool:
            initialize_registry_from_gcs(args.index_bucket_name, config, db_pool, dry_run=args.gcs_populate_dry_run,
                                         use_gcs_indexes_file=args.use_gcs_indexes_file,
                                         write_gcs_indexes_file=args.write_gcs_indexes_file,
                                         gcs_indexes_file_name=args.gcs_indexes_file_name,
                                         only_add_missing=args.only_add_missing,
                                         only_remove_extraneous=args.only_remove_extraneous)

    if args.add_index:
        with ReadWriteDbPool(config) as db_pool:
            add_index_from_path(args.add_index, config, db_pool)

    if args.defaults_from_config:
        indexes = default_indexes_from_config(json.loads(args.defaults_from_config))
        with ReadWriteDbPool(config) as db_pool:
            add_indexes(indexes, config, db_pool, only_add_missing=True)

    if args.gaps:
        starttime = datetime.datetime.strptime(args.start, "%Y-%m-%dT%H:%M:%SZ").replace(tzinfo=pytz.UTC)
        endtime = datetime.datetime.strptime(args.end, "%Y-%m-%dT%H:%M:%SZ").replace(tzinfo=pytz.UTC)

        with ReadWriteDbPool(config) as db_pool:
            org_gaps = find_gaps(config, db_pool, starttime, endtime, args.min_gap)
            print(json.dumps(org_gaps))


if __name__ == "__main__":
    sys.exit(main())
