import datetime
import logging
import os

from index_registry.engine.index_engine import IndexEngine, DISABLE_NON_VDR_INDEXES_ENV
from index_registry.engine.searcher_engine import SearcherEngine
from index_registry.utils.db import ReadWriteDbPool
from index_registry.utils.config import Config

from pscr_common.utils import get_utcnow_with_tzinfo
from pscr_common.log import setup_logging, clear_logging_handlers

_logger = logging.getLogger(__name__)


def purge(searcher_engine, index_engine, now):
    # Remove searchers that have expired 24 hours or more ago
    searcher_delete_count = 0
    try:
        searcher_delete_count = searcher_engine.purge_searchers(now - datetime.timedelta(hours=24))
        _logger.info("Deleted %s entries from searchers", searcher_delete_count)
    except:
        _logger.exception("Error deleting entries from searchers")

    # Remove indexes with date_removed greater than `PURGE_DATE_REMOVED_AGE_DAYS` days ago
    index_delete_count = 0
    try:
        purge_date_removed_age_days = int(os.environ.get("PURGE_DATE_REMOVED_AGE_DAYS", 10))
        index_delete_count = index_engine.purge_indexes(now - datetime.timedelta(days=purge_date_removed_age_days))
        _logger.info("Deleted %s entries from indexes", index_delete_count)
    except:
        _logger.exception("Error deleting entries from indexes")

    expired_count = 0
    try:
        # Expire indexes (set date_removed to now) for any indexes whose expiration_date has passed
        expired_count = index_engine.expire_indexes_with_expiration_date(now)
        _logger.info("Expired %s entries from indexes containing expiration date", expired_count)
    except:
        _logger.exception("Error expiring entries from indexes containing expiration date")

    if os.environ.get(DISABLE_NON_VDR_INDEXES_ENV, "false").lower() == "true":
        try:
            expired_non_vdr = index_engine.expire_indexes({}, {'default': 0, 'longterm': 0}, now)
            if expired_non_vdr > 0:
                _logger.warning("Immediately expired %s non-VDR indexes in a VDR-only environment", expired_non_vdr)
                expired_count += expired_non_vdr
        except:
            _logger.exception("Error expiring entries from indexes not containing expiration date")

    expired_stage2_count = 0
    try:
        # Expire old merge placeholders that haven't been written in a while, since those are not handled above
        expired_stage2_count = index_engine.expire_stage_2_placeholders(now)
        _logger.info("Expired %s old merge placeholders", expired_stage2_count)
    except:
        _logger.exception("Error expiring old merge placeholders")

    stale_org_expired_indexes = 0
    if os.environ.get("EXPIRE_STALE_ORGS", "false") == "true":
        try:
            stale_org_expired_indexes = index_engine.expire_stale_orgs(now)
            _logger.info("Expired %s entries from stale orgs", stale_org_expired_indexes)
        except:
            _logger.exception("Error expiring entries stale orgs")

    return {"deleted_searchers": searcher_delete_count,
            "deleted_leases": 0,
            "deleted_indexes": index_delete_count,
            "expired_indexes": expired_count,
            "expired_stale_merge_placeholders": expired_stage2_count,
            "expired_stale_org_indexes": stale_org_expired_indexes}

def main():
    config = Config.create()

    clear_logging_handlers()
    setup_logging(
        level=config.get("log", "level"),
        svc_version=os.environ.get("CONTAINERVERSION", "UNKNOWN"),
        env_id=os.environ.get("ENV_ID", "UNKNOWN"),
        svc_id="TODO",
    )

    db_pool = ReadWriteDbPool(config)
    now = get_utcnow_with_tzinfo()
    searcher_engine = SearcherEngine(config, db_pool=db_pool)
    index_engine = IndexEngine(os.environ.get("TRIGGER_SUFFIX", ".trigger"), config, db_pool=db_pool)

    res = purge(searcher_engine, index_engine, now)
    _logger.info(f"SQL purge results: {res}")
