import boto3
import logging
from dataclasses import dataclass

from index_registry.utils.config import Config
from index_registry.utils.exp_backoff import log_and_sleep

from botocore.exceptions import ClientError

_logger = logging.getLogger(__name__)

SEPARATOR = "_"
SQS_SERVICE = "rmqquery"

# TODO: This should go away when we clean up and do dev work for RabbitMQ
SNS_SERVICE = "sqsquery"


@dataclass
class TopicAttributes:
    subscriptions_created: int
    subscriptions_pending: int
    subscriptions_deleted: int


class SQSManager(object):

    def __init__(self, config: Config):
        self.config = config
        self.sqs_client = boto3.client('sqs', region_name=self.config.get('aws', 'default_region'))
        self.sns_client = boto3.client('sns', region_name=self.config.get('aws', 'default_region'))
        self.env = self.config.get('env', 'id')

    def delete_queue(self, queue_url):
        _logger.debug(f"Deleting SQS Queue {queue_url}")
        retries = 0
        while True:
            try:
                self.sqs_client.delete_queue(QueueUrl=queue_url)
            except ClientError as e:
                if e.response["Error"]["Code"] == "AWS.SimpleQueueService.NonExistentQueue":
                    _logger.warning(f"Tried to delete non existent queue {queue_url}")
                    return
                else:
                    raise
            except self.sqs_client.exceptions.InternalErrorException as e:
                log_and_sleep(str(e), retries, max_sleep=60)
            else:
                break
            retries += 1

    def _queues_by_prefix(self, prefix):
        kwargs = {'MaxResults': 1000,
                  'QueueNamePrefix': prefix}
        _logger.debug(f"Getting queues with prefix: {prefix}")
        while True:
            retries = 0
            while True:
                try:
                    queues = self.sqs_client.list_queues(**kwargs)
                except ClientError as e:
                    log_and_sleep(str(e), retries, max_sleep=60)
                except self.sqs_client.exceptions.InternalErrorException as e:
                    log_and_sleep(str(e), retries, max_sleep=60)
                else:
                    break
                retries += 1
            if 'QueueUrls' in queues:
                for queue_url in queues['QueueUrls']:
                    yield queue_url
                try:
                    kwargs['NextToken'] = queues['NextToken']
                except KeyError:
                    break
            else:
                _logger.warning(f"No queues found with prefix {prefix}")
                break

    @staticmethod
    def get_queue_from_url(queue_url):
        return queue_url.split('/')[-1]

    def queues(self):
        prefix = get_sqs_prefix(self.env)
        return self._queues_by_prefix(prefix)

    def queues_by_type(self, index_type):
        prefix = get_sqs_prefix((self.env)) + SEPARATOR + index_type
        return self._queues_by_prefix(prefix)

    def unsubscribe_queue(self, subscription_arn):
        if not subscription_arn:
            _logger.warning(f"No subscription arn")
        else:
            retries = 0
            while True:
                try:
                    self.sns_client.unsubscribe(SubscriptionArn=subscription_arn)
                except self.sns_client.exceptions.InternalErrorException as e:
                    log_and_sleep(str(e), retries, max_sleep=60)
                else:
                    break
                retries += 1

    def queue_empty(self, queue_name) -> bool:
        retries = 0
        while True:
            try:
                queue_url = self.sqs_client.get_queue_url(QueueName=queue_name)['QueueUrl']
            except ClientError as e:
                if e.response["Error"]["Code"] == "AWS.SimpleQueueService.NonExistentQueue":
                    _logger.warning(f"Queue {queue_name} does not exist returning True for empty")
                    return True
                else:
                    raise
            except self.sqs_client.exceptions.InternalErrorException as e:
                log_and_sleep(str(e), retries, max_sleep=60)
            else:
                break
            retries += 1
        retries = 0
        while True:
            try:
                attributes = self.sqs_client.get_queue_attributes(QueueUrl=queue_url,
                                                                  AttributeNames=['ApproximateNumberOfMessages'])
            except self.sqs_client.exceptions.InternalErrorException as e:
                log_and_sleep(str(e), retries, max_sleep=60)
            else:
                break
            retries += 1
        if int(attributes['Attributes']['ApproximateNumberOfMessages']) > 0:
            return False
        return True

    def delete_topic(self, topic_arn):
        # Note delete_topic is idempotent so we don't need to check for errors about topic not existing.
        self.sns_client.delete_topic(TopicArn=topic_arn)

    def get_topics(self):
        kwargs = dict()
        while True:
            retries = 0
            while True:
                _logger.debug("Calling list_topics")
                try:
                    topics = self.sns_client.list_topics(**kwargs)
                except ClientError as e:
                    log_and_sleep(str(e), retries, max_sleep=60)
                except self.sns_client.exceptions.InternalErrorException as e:
                    log_and_sleep(str(e), retries, max_sleep=60)
                else:
                    break
                retries += 1
            if 'Topics' in topics:
                for topic in topics['Topics']:
                    topic_name = topic['TopicArn'].split(':')[-1]
                    if topic_name.startswith(get_topic_prefix(self.env)):
                        yield topic
                try:
                    kwargs['NextToken'] = topics['NextToken']
                except KeyError:
                    break
            else:
                _logger.warning(f"No Topics Returned")
                break
        return

    def get_topic_attributes(self, topic_arn):
        retries = 0
        while True:
            try:
                ret = self.sns_client.get_topic_attributes(TopicArn=topic_arn)
            except self.sns_client.exceptions.InternalErrorException as e:
                log_and_sleep(str(e), retries, max_sleep=60)
            else:
                return TopicAttributes(subscriptions_created=int(ret['Attributes']['SubscriptionsConfirmed']),
                                       subscriptions_pending=int(ret['Attributes']['SubscriptionsPending']),
                                       subscriptions_deleted=int(ret['Attributes']['SubscriptionsDeleted']))
            retries += 1

    def get_topic_subscriptions(self):
        for topic in self.get_topics():
            attributes = self.sns_client.get_topic_attributes(TopicArn=topic['TopicArn'])
            if int(attributes['Attributes']['SubscriptionsConfirmed']) > 0:
                kwargs = {'TopicArn': topic['TopicArn']}
                while True:
                    retries = 0
                    while True:
                        _logger.debug(f"Getting subscriptions for {topic['TopicArn']}")
                        try:
                            ret = self.sns_client.list_subscriptions_by_topic(**kwargs)
                        except ClientError as e:
                            log_and_sleep(str(e), retries, max_sleep=60)
                        except self.sns_client.exceptions.InternalErrorException as e:
                            log_and_sleep(str(e), retries, max_sleep=60)
                        else:
                            break
                        retries += 1
                    for subscription in ret['Subscriptions']:
                        yield subscription
                    try:
                        kwargs['NextToken'] = ret['NextToken']
                    except KeyError:
                        break

    def queue_exists(self, queue_name) -> bool:
        try:
            self.sqs_client.get_queue_url(QueueName=queue_name)['QueueUrl']
        except ClientError as e:
            if e.response["Error"]["Code"] == "AWS.SimpleQueueService.NonExistentQueue":
                _logger.debug(f"Queue {queue_name} does not exist")
                return False
            else:
                raise
        return True

    def get_queue_create_timestamp(self, queue_url):
        try:
            ret = self.sqs_client.get_queue_attributes(QueueUrl=queue_url, AttributeNames=['CreatedTimestamp'])
        except ClientError as e:
            if e.response["Error"]["Code"] == "AWS.SimpleQueueService.NonExistentQueue":
                _logger.debug(f"Queue {queue_url} does not exist")
                return None
            else:
                raise
        return int(ret['Attributes']['CreatedTimestamp'])


def get_sqs_prefix(env):
    return env + SEPARATOR + SQS_SERVICE


# TODO: This should go away when we clean up and do dev work for RabbitMQ
def get_topic_prefix(env):
    return env + SEPARATOR + SNS_SERVICE


def get_consumer_name(properties, env):
    return get_sqs_prefix(env) + SEPARATOR + \
           properties['type'] + SEPARATOR + \
           properties['org'] + SEPARATOR + \
           properties['class'] + SEPARATOR + \
           properties['partition'] + SEPARATOR + \
           properties['tstamp']


def get_org_from_consumer_name(consumer_name):
    org_key = consumer_name.split(SEPARATOR)[3]
    #  Special case for where the org_key is a leading underscore
    if org_key == "":
        org_key = '_' + consumer_name.split(SEPARATOR)[4]
    return org_key


# TODO: This should go away when we clean up and do dev work for RabbitMQ
def get_topic_name(env, org):
    return get_sqs_prefix(env) + SEPARATOR + org




