import os
import logging

from fastapi import FastAPI
from pscr_common.metrics import PrometheusRegistry

from index_registry.engine.index_engine import IndexEngine
from index_registry.engine.searcher_engine import SearcherEngine
from index_registry.engine.org_sqs_engine import OrgSqsEngine
from index_registry.mongodb.leases_db_interface import DbManager as LeasesDbManager
from index_registry.utils.db import ReadWriteDbPool
from index_registry.utils.config import Config
from index_registry.utils.metrics_reporter import start_reporter
from index_registry.web.registry_routes import build_routes
from index_registry.web.middleware import build_metrics_middleware

from pscr_common.log import setup_logging, clear_logging_handlers

logging.getLogger("pymongo").setLevel(logging.CRITICAL)


def build_app() -> FastAPI:
    """Initialise the fastAPI app with all its dependencies
    """
    config = Config.create()

    clear_logging_handlers()
    if os.environ.get("PSCR_LOGGING", "true").lower() == "true":
        setup_logging(
            level=config.get("log", "level"),
            svc_version=os.environ.get("CONTAINERVERSION", "UNKNOWN"),
            env_id=os.environ.get("ENV_ID", "UNKNOWN"),
            svc_id="TODO",
        )
    else:
        logging.basicConfig(level=os.environ.get("LOG_LEVEL", "DEBUG"))

    _logger = logging.getLogger(__name__)

    metrics_registry = PrometheusRegistry()
    start_reporter(metrics_registry)

    db_pool = ReadWriteDbPool(config)
    leases_db_manager = LeasesDbManager(config)

    index_engine = IndexEngine(os.environ.get("TRIGGER_SUFFIX", ".trigger"), config, db_pool=db_pool)
    searcher_engine = SearcherEngine(config=config, db_pool=db_pool)

    org_sqs_engine = OrgSqsEngine(config)

    _app = FastAPI()

    build_metrics_middleware(_app, metrics_registry)
    build_routes(_app, config, leases_db_manager, index_engine,
                 searcher_engine, org_sqs_engine)

    return _app


if os.getenv("REGISTRY_API_RUN", "false").lower() == "true":
    app = build_app()
