import time
import logging

from fastapi import Request
from pscr_common.metrics import PrometheusRegistry

_logger = logging.getLogger(__name__)


def get_route_metric_label(request: Request) -> str:
    if not request or not request.url:
        return "unknown_route"
    parts = [p.strip("/") for p in request.url.path.split("/") if len(p) > 0]
    parts = [
        p.strip("{}").replace("_", "").upper() if p.startswith("{") else p
        for p in parts
    ]  # Upper variable names
    return "_".join(parts) + "." + request.method.lower()


def build_metrics_middleware(app, metrics_registry: PrometheusRegistry):
    """This middleware factory creates middleware that reports metrics for each request"""

    @app.middleware("http")
    async def get_request_metrics(request: Request, call_next):

        request_metric_prefix = f"requests.{get_route_metric_label(request)}"
        response_status = 200

        try:
            with metrics_registry.timer(request_metric_prefix + ".time").time():
                response = await call_next(request)
                response_status = response.status_code

                return response
        except Exception as ex:
            _logger.exception(f"Error during request {str(request)}: {str(ex)}")
            response_status = 500
        finally:
            metrics_registry.meter(
                f"{request_metric_prefix}.status.{response_status}"
            ).mark()
